﻿using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;

// VMdemoWinForms
// Example of an approach to ViewModel for scientific software, using WPF
// Released for educational purposes, without warranty or restrictions.
// Compliments of:
//    Michael A. Covington, Ph.D., Consultant
//    Covington Innovations
//    http://www.covingtoninnovations.com/michael

// Key idea:
//
// The purpose of the ViewModel paradigm is to restrict interaction between
// the user interface (Form1, the "view") and the computation engine 
// (the "model" in older nomenclature).
//
// The Form1ViewModel object contains everything that Form1 and
// the computation engine need to communicate with each other.  They do
// not communicate with each other directly; they use it as an intermediary.
//
// Thus, the programmer can easily manage all possible interactions between
// the two.  The user interface code does not become cluttered with heterogeneous
// material interacting in various ways with the computation or with the GUI.
//
// If the user interface were to change completely (e.g., to WPF or to
// a command line interface), Form1ViewModel would have to be rewritten
// but the computation engine would not.  Conversely, major changes in the
// computation not involving a change in the user experience might involve
// changes to Form1ViewModel but not to Form1.
//
// In a WinForms application, communication between Form1 and Form1ViewModel
// takes place by directly retrieving or setting properties of Form1 components.
// In many cases, the setters have side effects (perform other manipulations of
// Form1).  Although limited data binding is available for some components,
// I have not used it here.
//
// NOTE: All components of Form1 referenced here must be declared public.
// This is done by changing "private" to "public" in the last few lines
// of Form1.Designer.cs.
//
// NOTE: Awkwardly, all communication between Form1ViewModel and Form1 is
// potentially cross-thread and requires Invoke or BeginInvoke.

// See also the comments in the other .cs files.

namespace VMdemoWinForms
{
    /// <summary>
    /// ViewModel for Form1 of this program. 
    /// If there were multiple windows, there would be a 
    /// ViewModel class for each, appropriately named.
    /// </summary>
    public class Form1ViewModel
    {

        #region Constructor (initialization)

        private Form1 form1;

        private Form1ViewModel() { }  // do not use argumentless constructor

        /// <summary>
        /// Create the ViewModel by passing it a reference to the Form1 to
        /// which it belongs, so that it can access that window directly in
        /// situations where data binding isn't sufficient.
        /// </summary>
        public Form1ViewModel(Form1 f)
        {
            form1 = f;
            _Running = false;

            Canceler = new CancellationTokenSource();
            // Initialized here so it will never be null,
            // but re-initialized before each use so it will be fresh
        }

        #endregion

        #region Canceler for signaling asynchronous processes to exit

        public CancellationTokenSource Canceler;

        #endregion

        #region Data displayed on the screen

        /// <summary>
        /// String holding the number of seconds to count (typed by the user).
        /// Getter retrieves it from the textbox; setter places it there.
        /// </summary>
        public string SecondsToCount
        {
            get
            {
                string result = "";
                form1.Invoke(new MethodInvoker(delegate()
                    {
                        result = form1.textBoxSeconds.Text;
                    }
                ));
                return result;
            }
            set
            {
                form1.Invoke(new MethodInvoker(delegate()
                    {
                        form1.textBoxSeconds.Text = value;
                    }
                ));
            }
        }

        /// <summary>
        /// String holding the displayed countdown.
        /// Bound to the textbox where it is displayed.
        /// </summary>
        public string DisplayedCountdown
        {
            get
            {
                string result = "";
                form1.Invoke(new MethodInvoker(delegate()
                {
                    result = form1.textBoxDisplayedCountdown.Text;
                }
                ));
                return result;
            }
            set
            {
                form1.Invoke(new MethodInvoker(delegate()
                {
                    form1.textBoxDisplayedCountdown.Text = value;
                }
                ));
            }
        }

        /// <summary>
        /// The second text box is a write-only log, like writing to a file,
        /// so instead of being data-bound to a string, it receives data through
        /// calls to this method.
        /// </summary>
        /// <param name="s">
        /// String to be written.  Be sure to include end-of-line mark if needed.
        /// </param>
        public void AppendToLog(string s)
        {
            // Any method in the ViewModel that directly manipulates the MainWindow
            // (not through binding) may be performing a cross-thread call.
            // Here's how it's done:

            form1.BeginInvoke(new MethodInvoker(delegate()
                    {
                        form1.textBoxLog.AppendText(s);
                        form1.textBoxLog.ScrollToCaret();
                    }
           ));
        }

        #endregion

        #region Program state indicators
#warning The program state indicators are what's not working here!
#warning Probably need an UpdateGui method.

        /// <summary>
        /// Read-only property indicating whether the computation is ready to start.
        /// True if Running is false and there is a valid integer in the SecondsToCount box.
        /// Bound to Start button's Enabled property.
        /// </summary>
        public bool Ready
        {
            get
            {
                int junk;
                return (Running == false && int.TryParse(SecondsToCount, out junk));
            }
        }

        /// <summary>
        /// Read-only property indicating whether the user interface can accept input.
        /// True if Running is false.
        /// Bound to Seconds To Count textbox's Enabled property.
        /// </summary>
        public bool NotRunning
        {
            get
            {
                return !Running;
            }
        }

        /// <summary>
        /// Property indicating whether the computation is currently in progress.
        /// Enabled property of the Cancel button is bound to this.
        /// Its setter also controls the cursor (arrow vs. wait cursor).
        /// </summary>
        public bool Running
        {
            get
            {
                return _Running;
            }
            set
            {
                _Running = value;
                UpdateGuiAppearance();
            }
        }

        private bool _Running;

        /// <summary>
        /// Method to update the appearance of the GUI -- enable or disable
        /// components and change cursors depending on the state of the
        /// computation.  This is less elegant than the way these things
        /// were elegantly broken out in the WPF version of the program.
        /// </summary>
        public void UpdateGuiAppearance()
        {
            // This is called by the setter of Running and 
            // by the TextChanged event of textBoxSeconds.

            // Run on form1's thread
            form1.BeginInvoke(new MethodInvoker(delegate()
            {
                // Enable buttonStart iff not running and there is a valid integer in SecondsToCount
                int junk;
                form1.buttonStart.Enabled = (_Running == false && int.TryParse(SecondsToCount, out junk));

                // Enable textBoxSeconds iff not running
                form1.textBoxSeconds.Enabled = (!_Running);

                // Enable buttonCancel iff running
                form1.buttonCancel.Enabled = _Running;

                // Change cursors depending on whether running
                if (_Running)
                {
                    form1.Cursor = Cursors.WaitCursor;
                    form1.textBoxSeconds.Cursor = Cursors.WaitCursor;
                    form1.textBoxDisplayedCountdown.Cursor = Cursors.WaitCursor;
                    form1.textBoxLog.Cursor = Cursors.WaitCursor;
                    form1.buttonCancel.Cursor = Cursors.Arrow;
                }
                else
                {
                    form1.Cursor = Cursors.Arrow;
                    form1.textBoxSeconds.Cursor = Cursors.IBeam;
                    form1.textBoxDisplayedCountdown.Cursor = Cursors.IBeam;
                    form1.textBoxLog.Cursor = Cursors.IBeam;
                    form1.buttonCancel.Cursor = Cursors.Arrow;
                }
            }
           ));
        }


        #endregion

        #region Methods to launch and cancel the computation

        /// <summary>
        /// Launch the computation in a concurrent process.
        /// Called by the event handler of the Start button.
        /// </summary>
        public async Task LaunchComputation()
        {
            Canceler = new CancellationTokenSource(); // fresh one for each run
            Running = true;
            await Task.Factory.StartNew( 
                () => Computation.Countdown(this) 
            );
            Running = false;
        }

        /// <summary>
        /// Since this computation does not need to end gracefully, we
        /// just kill the process and then reset some values.
        /// We could use a CancellationTokenSource if it needed to break
        /// out of a main loop gracefully.
        /// Called by the event handler of the Cancel button.
        /// </summary>
        public void CancelComputation()
        {
            Canceler.Cancel();
        }

        #endregion
    }
}


